with Adagraph;
with Common;
with Room;

package robot is
   -- Robot rules
   --    (1)  Initially the robot faces north, arms down, and its light off
   --    (2)  When the robot's light is on, it ignores all command, except light
   type robot_type is limited private;
   
   -- This exception is raised whenever the robot encounters and instruction it
   -- does not recognize
   instruction_error : exception;
   
   -- load allows the robot program to be downloaded into the robot
   -- memory for execution.  Valid instructions are step, turn, raise, lower,
   -- sense, goto, light, and stop.  Only the goto instruction has an operand,
   -- which is the instruction number of the next instruction to execute.
   -- Invalid instructions are ignored.
   procedure load (R : in out robot_type);
   
   -- fetches the next instruction for execution
   procedure fetch (R : in out robot_type);
   
   -- execute causes the robot to execut the next instruction retrieved by fetch
   -- according to the semantics indicated in the procedure comments below.
   procedure execute (R : in out robot_type);
   
   -- draws the robot using Adagraph
   procedure draw(R : in robot_type);
   
   procedure place (R : in out robot_type; X : in Positive; Y : in Positive);
   procedure location (R : in robot_type; X : out Positive; Y : out Positive);
   
   function end_of_program (R : robot_type) return Boolean;
   
   private
   
   -- Step causes the robot to take one step forward, if possible.  If the robot
   -- cannot step forward, it remains where it is.  
   -- The robot cannot step forward if:
   --    (1) it is next to a wall and facing the wall
   --    (2) it is one step from the wall and facing the wall with its arms raised
   procedure step (R : in out robot_type);
   
   -- Turn causes the robot to turn to the right 90 degrees, if possible.
   -- The robot cannot turn if:
   --     (1) it has a wall on its immediate right and its arms are raised
   procedure turn (R : in out robot_type);
   
   -- Raise_arms causes the robot to raise its arms, if possible.  If the robot
   -- cannot raise its arms, then it turns on its warning light.
   -- The robot cannot raise its arms if:
   --    (1) it is next to a wall and facing the wall
   procedure raise_arms (R : in out robot_type);
   
   -- Lower_arms causes the robot to lower its arms.
   procedure lower_arms (R : in out robot_type);
   
   -- Sense causes the robot to discover a wall one step in front of the robot,
   -- if the robot has its arms raised.  If a wall is discovered, then the robot
   -- turns on its warning light.
   procedure sense (R : in out robot_type);
   
   -- Light causes the robot to turn off its warning light.
   procedure light (R : in out robot_type);
   
   -- Jump sets the robot's program counter to the operand value
   -- in the gotoi command
   procedure jump (R : in out robot_type); 
   
   num_instructions      : constant Positive := 8;
   max_memory_location   : constant Positive := 100;   

   -- The sizes selected assume a room tile size based on a multiple 6
   Body_Radius           : constant Positive := Room.Tile_Size / 2;
   Light_Radius          : constant Positive := Room.Tile_Size / 6;
   Arm_Length            : constant Positive := Room.Tile_Size / 1;
   Arm_Width             : constant Positive := Room.Tile_Size / 6;
   Body_Color            : constant Adagraph.Extended_Color_Type := Adagraph.Blue;
   Arm_Color             : constant Adagraph.Extended_Color_Type := Adagraph.Green;
   Light_Color           : constant Adagraph.Extended_Color_Type := Adagraph.Red;
   
   type instruction_type is (stepi, turni, raisei, loweri, sensei, gotoi, lighti, stopi);
   subtype operand_type  is Natural range 0..max_memory_location - 1;
   
   type inst_rec_type    is record
      instruction : instruction_type := stopi;
      operand     : operand_type := 0;
   end record;
   
   type memory_type      is array (0..max_memory_location - 1) of inst_rec_type;

   type arm_type         is (up, down);
   type light_type       is (on, off);
   
   type robot_type is record
      memory    : memory_type;
      X, Y      : Natural;
      arms      : arm_type              := down;
      light     : light_type            := off;
      direction : Common.Direction_Type := Common.north;
      pc        : operand_type := 0;
      jump      : Boolean := True;
   end record;
end robot;