with Ada.Text_IO;

package body robotADT is

   -- Integer IO package for output to strings in the function "instruction"
   package Int_IO  is new Ada.Text_IO.Integer_IO(Num => Integer);

   ------------------
   -- load_program --
   ------------------

   procedure load_program (R : in out robot_type; Filename : in String) is
      File        : Ada.Text_IO.File_Type;
      I           : Integer := 0;
      opcode      : String(1..4);
      operand     : Natural;
      binop       : Natural;
   begin
      Ada.Text_IO.Open(Name => Filename(1..Filename'Length), 
         Mode => Ada.Text_IO.In_File, File => File);
      
      while I <= address_subtype'last and then not Ada.Text_IO.End_Of_File (File) loop
         begin -- exeception handler for file input
         
            -- Get a robot instruction 
            Ada.Text_IO.Get(File => File, Item => opcode);

            case opcode(1) is
               when 't' | 'T' => R.memory(I).opcode := turni;
               when 'r' | 'R' => R.memory(I).opcode := raisei;
               when 'g' | 'G' => 
                  R.memory(I).opcode := gotoi;
                  
                  -- Only the gotoi instruction has an operand   
                  Int_IO.Get(File => File, Item => operand);
                  
                  -- convert the binary repersentation to a decimal number
                  binop := 0;
                  for bit in reverse 0 .. numabits - 1 loop
                     binop := binop + operand / 10**bit * 2**bit;
                     if operand >= 10**bit then operand := operand - 10**bit; end if;
                  end loop;
                  
                  R.Memory(I).operand := binop;
               when 'l' | 'L' =>
                  if opcode(2) = 'o' or opcode(2) = 'O' then
                     R.memory(I).opcode := loweri;
                  else
                     R.memory(I).opcode := lighti;
                  end if;
               when 's' | 'S' =>
                  if opcode(3) = 'e' or opcode(3) = 'E' then
                     R.memory(I).opcode := stepi;
                  elsif opcode(3) = 'n' or opcode(3) = 'N' then
                     R.memory(I).opcode := sensei;
                  else
                     R.memory(I).opcode := stopi;
                  end if;
               when others =>
                  raise Ada.Text_IO.Data_Error;
            end case;

            -- Skip the remainder (if any) of the record
            Ada.Text_IO.Skip_Line(File => File);

            I := I + 1;
         exception -- handler for file input
            when Ada.Text_IO.End_Error => 
               null;
            when Ada.Text_IO.Data_Error =>
               Ada.Text_IO.New_Line;
               Ada.Text_IO.Put (Item => "Unrecognized data in line:");
               Ada.Text_IO.Set_Col(30);
               Int_IO.Put (Item => I);
               Ada.Text_IO.Put (Item => ".  Discarding this line.");
               Ada.Text_IO.New_Line;
               Ada.Text_IO.Skip_Line (File => File);
               delay 0.5;
            when Constraint_Error =>
               Ada.Text_IO.New_Line;
               Ada.Text_IO.Put (Item => "Invalid memory address in line:");
               Ada.Text_IO.Set_Col(40);
               Int_IO.Put (Item => I);
               Ada.Text_IO.Put (Item => ".  Discarding this line.");
               Ada.Text_IO.New_Line;
               Ada.Text_IO.Skip_Line (File => File);
               delay 0.5;
         end; -- exception handler for file input
      end loop;
      
      Ada.Text_IO.Close (File => File);
      
      -- Fill up the remainder of the robot's memory with noopis
      for J in I..address_subtype'last loop
         R.memory(J).opcode  := noopi;
         R.memory(J).operand := 0;
      end loop;
      
      -- reset the program counter to the beginning of the newly loaded program
      R.PC := 0;

      exception
         when Ada.Text_IO.Use_Error | Ada.Text_IO.Name_Error =>
            Ada.Text_IO.Put (Item => "Error opening input file.  File not read.");
            Ada.Text_IO.New_Line;
   end load_program;

   -------------------
   -- store_program --
   -------------------

   procedure store_program (R : in robot_type; Filename : in String) is
      File        : Ada.Text_IO.File_Type;
      I           : Integer := 0;
      inst_string : String(1..5+numabits);
   begin
      Ada.Text_IO.Create(Name => Filename(1..Filename'Length), 
         Mode => Ada.Text_IO.Out_File, File => File);

      while I <= address_subtype'last loop
         inst_string := instruction(R.memory(I).opcode, R.memory(I).operand);
         case R.memory(I).opcode is
            when gotoi  => Ada.Text_IO.Put (File => File, Item => inst_string);
            when others => Ada.Text_IO.Put (File => File, Item => inst_string(1..5));
         end case;
         Ada.Text_IO.New_Line(File => File);
         I := I + 1;
      end loop;
      
      Ada.Text_IO.Close (File => File);

      exception
         when Ada.Text_IO.Use_Error | Ada.Text_IO.Status_Error | Ada.Text_IO.Device_Error =>
            Ada.Text_IO.Put (Item => "Error create output file.  File not created.");
            Ada.Text_IO.New_Line;
   end store_program;

   -----------------------
   -- fetch_and_execute --
   -----------------------

   procedure fetch_and_execute (R : in out robot_type;
      Dynamic_Behavior : in Dynamic_Behavior_Pointer) is
      opcode : opcode_type := R.Memory(R.PC).opcode;
   begin
      case opcode is
         when stepi  =>
            step (R => R);
         when turni  =>
            turn (R => R);
         when raisei =>
            raise_arms (R => R);
         when loweri =>
            lower_arms (R => R);
         when sensei =>
            sense (R => R, Dynamic_Behavior => Dynamic_Behavior);
         when gotoi  =>
            jump (R => R);
         when lighti =>
            light (R);
         when stopi | noopi =>
            null;
      end case;

      begin
         case opcode is
            when stepi | turni | raisei | loweri | sensei | lighti =>
               R.pc := R.pc + 1;
            when gotoi|stopi|noopi =>
               if R.light = on then
                  R.pc := R.pc + 1;
               end if;
         end case;
      exception
         when Constraint_Error =>
            raise Memory_Error;
      end;
   end fetch_and_execute;

   --------------------
   -- end_of_program --
   --------------------

   function end_of_program (R : robot_type) return Boolean is
   begin
      if (R.Light=off and (R.memory(R.pc).opcode = stopi or R.memory(R.PC).opcode = noopi)) or 
         (R.X<1 or R.X>RoomADT.Size(R.room.all)) or (R.Y<1 or R.Y>RoomADT.Size(R.room.all)) then
         return True;
      else
         return False;
      end if;
   end end_of_program;

   ------------------
   -- get_location --
   ------------------

   procedure get_location
     (R : in     robot_type;
      X : out Natural;
      Y : out Natural)
   is
   begin
      X      := R.X;
      Y      := R.Y;
   end get_location;

   ------------------
   -- set_location --
   ------------------

   procedure set_location
     (R    : in out robot_type;
      X    : in     Positive;
      Y    : in     Positive;
      room : in     RoomADT.room_type_ptr)
   is
      height : Positive := RoomADT.Height(room.all);
      width  : Positive := RoomADT.Width (room.all);
   begin
      if X > RoomADT.width(room.all) or Y > RoomADT.Height(room.all) then
         Ada.Text_IO.Put (Item => "set_location called with invalid position for robot");
         Ada.Text_IO.New_Line;
      else
         R.room := room;
         case R.direction is
            when north =>
               if (Y = height and R.arms = up and not RoomADT.Is_Door(R.room.all, X, Y+1)) then
                  raise instruction_error;
               end if;
            when south =>
               if (Y = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, X, 0)) then
                  raise instruction_error;
               end if;
            when east  =>
               if (X = width and R.arms = up and not RoomADT.Is_Door(R.room.all, X+1, Y)) then
                  raise instruction_error;
               end if;
            when west  =>
               if (X = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, 0, Y)) then
                  raise instruction_error;
               end if;
         end case;
         R.X    := X;
         R.Y    := Y;
      end if;
   end set_location;

   --------------
   -- get_arms --
   --------------

   procedure get_arms (R : in     robot_type; state : out arm_type) is
   begin
      state := R.arms;
   end get_arms;

   --------------
   -- set_arms --
   --------------

   procedure set_arms (R : in out robot_type; state : in  arm_type) is
   begin
      if state=up then
         raise_arms(R);
      else
         R.arms := down;
      end if;
   end set_arms;

   ---------------
   -- get_light --
   ---------------

   procedure get_light (R : in     robot_type; state : out light_type) is
   begin
      state := R.light;
   end get_light;

   ---------------
   -- set_light --
   ---------------

   procedure set_light (R : in out robot_type; state : in  light_type) is
   begin
      R.light := state;
   end set_light;

   -------------------
   -- get_direction --
   -------------------

   procedure get_direction
     (R : in     robot_type;
      state : out direction_type)
   is
   begin
      state := R.direction;
   end get_direction;

   -------------------
   -- set_direction --
   -------------------

   procedure set_direction
     (R : in out robot_type;
      state : in  direction_type)
   is
      height : Positive := RoomADT.Height(R.room.all);
      width  : Positive := RoomADT.Width (R.room.all);
   begin
      case state is
         when north =>
            if not (R.Y = height and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X, R.Y+1)) then
               R.direction := north;
            else
               raise instruction_error;
            end if;
         when south =>
            if not (R.Y = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X, 0)) then
               R.direction := south;
            else
               raise instruction_error;
            end if;
         when east  =>
            if not (R.X = width and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X+1, R.Y)) then
               R.direction := east;
            else
               raise instruction_error;
            end if;
         when west  =>
            if not (R.X = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, 0, R.Y)) then
               R.direction := west;
            else
               raise instruction_error;
            end if;
      end case;
   end set_direction;

   ------------
   -- get_PC --
   ------------

   procedure get_PC (R : in     robot_type; state : out address_subtype) is
   begin
      state := R.PC;
   end get_PC;

   ------------
   -- set_PC --
   ------------

   procedure set_PC (R : in out robot_type; state : in  address_subtype) is
   begin
      R.PC := state;
   end set_PC;

   -----------------
   -- get_address --
   -----------------

   function get_address
     (R : in     robot_type;
      loc : address_subtype)
      return instruction_subtype
   is
   begin
      return instruction(opcode => R.memory(loc).opcode, operand => R.memory(loc).operand);
   end get_address;

   -----------------
   -- get_address --
   -----------------

   function get_address
     (R : in     robot_type;
      loc : address_subtype)
      return String
   is
   begin
      return instruction(opcode => R.memory(loc).opcode, operand => R.memory(loc).operand);
   end get_address;

   -----------------
   -- set_address --
   -----------------

   procedure set_address
     (R : in out robot_type;
      loc : address_subtype;
      state : in  instruction_subtype)
   is
   begin
      case state is
         when stepcode..stepcode+oprdcode =>
            R.memory(loc).opcode := stepi;
            R.memory(loc).operand := state - stepcode;
         when turncode..turncode+oprdcode =>
            R.memory(loc).opcode := turni;
            R.memory(loc).operand := state - turncode;
         when raiscode..raiscode+oprdcode =>
            R.memory(loc).opcode := raisei;
            R.memory(loc).operand := state - raiscode;
         when lowrcode..lowrcode+oprdcode =>
            R.memory(loc).opcode := loweri;
            R.memory(loc).operand := state - lowrcode;
         when senscode..senscode+oprdcode =>
            R.memory(loc).opcode := sensei;
            R.memory(loc).operand := state - senscode;
         when litecode..litecode+oprdcode =>
            R.memory(loc).opcode := lighti;
            R.memory(loc).operand := state - litecode;
         when stopcode..stopcode+oprdcode =>
            R.memory(loc).opcode := stopi;
            R.memory(loc).operand := state - stopcode;
         when gotocode..gotocode+oprdcode =>
            R.memory(loc).opcode  := gotoi;
            R.memory(loc).operand := state - gotocode;
         when others =>
            raise instruction_error;
      end case;
   end set_address;

   ----------
   -- step --
   ----------

   procedure step (R : in out robot_type) is
      door   : Boolean;
      height : Positive := RoomADT.Height(R.room.all);
      width  : Positive := RoomADT.Width (R.room.all);
   begin
      case R.direction is
         when north =>
            door := RoomADT.Is_Door (R.room.all, R.X, height+1);
            if not ((R.Y = height and not door) or 
                    (R.Y + 1 = height and R.arms = up and not door)) then
               R.Y := R.Y + 1;
            end if;
         when south =>
            door := RoomADT.Is_Door (R.room.all, R.X, 0);
            if not ((R.Y = 1 and not door) or (R.Y - 1 = 1 and R.arms = up and not door)) then
               R.Y := R.Y - 1;
            end if;
         when east  =>
            door := RoomADT.Is_Door (R.room.all, width+1, R.Y);
            if not ((R.X = width and not door) or 
                    (R.X + 1 = width and R.arms = up and not door)) then
               R.X := R.X + 1;
            end if;
         when west  =>
            door := RoomADT.Is_Door (R.room.all, 0, R.Y);
            if not ((R.X = 1 and not door) or (R.X - 1 = 1 and R.arms = up and not door)) then
               R.X := R.X - 1;
            end if;
      end case;
   end step;

   ----------
   -- turn --
   ----------

   procedure turn (R : in out robot_type) is
      height : Positive := RoomADT.Height(R.room.all);
      width  : Positive := RoomADT.Width (R.room.all);
   begin
      if R.light = off then
         case R.direction is
         when north =>
            if not (R.X = width and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X+1, R.Y)) then
               R.direction := east;
            end if;
         when south =>
            if not (R.X = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, 0, R.Y)) then
               R.direction := west;
            end if;
         when east  =>
            if not (R.Y = 1 and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X, 0)) then
               R.direction := south;
            end if;
         when west  =>
            if not (R.Y = height and R.arms = up and not RoomADT.Is_Door(R.room.all, R.X, R.Y+1)) then
               R.direction := north;
            end if;
         end case;
      end if;
   end turn;

   ----------------
   -- raise_arms --
   ----------------

   procedure raise_arms (R : in out robot_type) is
      height : Positive := RoomADT.Height(R.room.all);
      width  : Positive := RoomADT.Width (R.room.all);
   begin
      if R.light = off then
         case R.direction is
            when North =>
               if R.Y = height and not RoomADT.Is_Door(R.room.all, R.X, R.Y+1) then
                  R.light := on;
               else
                  R.arms := up;
               end if;
            when South =>
               if R.Y = 1 and not RoomADT.Is_Door(R.room.all, R.X, 0) then
                  R.light := on;
               else
                  R.arms := up;
               end if;
            when East  =>
               if R.X = width and not RoomADT.Is_Door(R.room.all, R.X+1, R.Y) then
                  R.light := on;
               else
                  R.arms := up;
               end if;
            when West  =>
               if R.X = 1 and not RoomADT.Is_Door(R.room.all, 0, R.Y) then
                  R.light := on;
               else
                  R.arms := up;
               end if;
         end case;
      end if;
   end raise_arms;

   ----------------
   -- lower_arms --
   ----------------

   procedure lower_arms (R : in out robot_type) is
   begin
      if R.light = off then
         R.arms := down;
      end if;
   end lower_arms;

   -----------
   -- sense --
   -----------

   procedure sense (R : in out robot_type;
      Dynamic_Behavior : in Dynamic_Behavior_Pointer) is
      height : Positive := RoomADT.Height(R.room.all);
      width  : Positive := RoomADT.Width (R.room.all);
   begin
      if R.arms = up and R.light = off then
         case R.direction is
            when north =>
               Dynamic_Behavior.all(R.X,R.Y+1);
               if R.Y + 1 = height and not RoomADT.Is_Door(R.room.all, R.X, R.Y+2) then
                  R.light := on;
               end if;
            when south =>
               Dynamic_Behavior.all(R.X,R.Y-1);
               if R.Y - 1 = 1 and not RoomADT.Is_Door(R.room.all, R.X, 0) then
                  R.light := on;
               end if;
            when east  =>
               Dynamic_Behavior.all(R.X+1,R.Y);
               if R.X + 1 = width and not RoomADT.Is_Door(R.room.all, R.X+2, R.Y) then
                  R.light := on;
               end if;
            when west  =>
               Dynamic_Behavior.all(R.X-1,R.Y);
               if R.X - 1 = 1 and not RoomADT.Is_Door(R.room.all, 0, R.Y) then
                  R.light := on;
               end if;
         end case;
      end if;
   end sense;

   ----------
   -- jump --
   ----------

   procedure jump (R : in out robot_type) is
   begin
      if R.light = off then
         R.PC := R.Memory(R.PC).operand;
      end if;
   end jump;

   -----------
   -- light --
   -----------

   procedure light (R : in out robot_type) is
   begin
      R.light := off;
   end light;
   
   function instruction (opcode : opcode_type; operand : address_subtype) 
      return instruction_subtype is
   begin
      case opcode is
         when stepi  => return stepcode + operand;
         when turni  => return turncode + operand;
         when raisei => return raiscode + operand;
         when loweri => return lowrcode + operand;
         when sensei => return senscode + operand;
         when gotoi  => return gotocode + operand;
         when lighti => return litecode + operand;
         when stopi | noopi  => return stopcode + operand;
      end case;            
   end instruction;
   
   function instruction (opcode : opcode_type; operand : address_subtype) return String is
      -- The string initialization here and the literal assignments below may change according
      -- to the number of bits used to address memory.  That is, if the robot's memory size
      -- changes, then the literal strings in this procedure may also need to change to remain
      -- compatible.
      inst_string : String(1..5+numabits) := (others => ' ');
      D : Natural;
      N : Natural := operand;
   begin
      inst_string(1..4) := "GOTO";
      case opcode is
         when stepi  => 
            inst_string(1..4) := "STEP";
         when turni  => 
            inst_string(1..4) := "TURN";
         when raisei => 
            inst_string(1..5) := "RAISE";
         when loweri => 
            inst_string(1..5) := "LOWER";
         when sensei => 
            inst_string(1..5) := "SENSE";
         when gotoi  => 
            for I in reverse 0..numabits-1 loop
               D := N / (2**I);
               Int_IO.Put (To => inst_string(5+numabits-I..5+numabits-I), Item => D);
               N := N - D * (2**I);
            end loop;
            return inst_string;
         when lighti => 
            inst_string(1..5) := "LIGHT";
         when stopi | noopi  => 
            inst_string(1..4) := "STOP";
      end case;  
      return inst_string;          
   end instruction;
   
   procedure Insert_Instruction(R : in out robot_type) is
      PC : Natural;
   begin
      Get_PC(R,PC);
      for I in Address_Subtype loop
         if R.Memory(I).opcode = gotoi then
            if R.Memory(I).operand >= PC and 
               R.Memory(I).operand < Address_Subtype'Last then
               R.Memory(I).operand := R.Memory(I).operand + 1;
            end if;
         end if;
      end loop;
      for I in reverse PC..Address_Subtype'Last-1 loop
         RobotADT.Set_Address(
            R     => R,
            loc   => I+1,
            state => Get_Address(R,I));
      end loop;
   end Insert_Instruction;
   
   procedure Delete_Instruction(R : in out robot_type) is
      PC : Natural;
   begin
      RobotADT.Get_PC(R,PC);
      for I in Address_Subtype loop
         if R.Memory(I).opcode = gotoi then
            if R.Memory(I).operand > PC then
               R.Memory(I).operand := R.Memory(I).operand - 1;
            end if;
         end if;
      end loop;
      for I in PC..Address_Subtype'Last-1 loop
         RobotADT.Set_Address(
            R     => R,
            loc   => I,
            state => Get_Address(R,I+1));
      end loop;
      RobotADT.Set_Address(
         R     => R,
         loc   => Address_Subtype'Last,
         state => StopCode);
   end Delete_Instruction;
end robotADT;