with RoomADT;

package robotADT is
   -- EXPORTED TYPES
   type    robot_type          is limited private;
   type    direction_type      is (north, south, east, west);
   type    arm_type            is (up, down);
   type    light_type          is (on, off);
   
   numabits : constant := 5;
   subtype address_subtype     is Natural range 0..2**numabits-1;
   subtype instruction_subtype is Natural range 0..255;
   type Dynamic_Behavior_Pointer is access procedure (X,Y : in Natural);

   -- EXPORTED EXCEPTIONS
   instruction_error : exception;  -- raised whenever robot encounters an unrecognized instruction
   memory_error      : exception;  -- raised when a reference occurs outside robot memory
   
   -- EXPORTED PROCEDURES/FUNCTIONS
   
   -- Downloads the robot program from the named file into the robot
   -- memory for execution.  Valid instructions are step, turn, raise, lower,
   -- sense, goto, light, and stop.  Only the goto instruction has an operand,
   -- which is the instruction number of the next instruction to execute.
   -- Invalid instructions are ignored.
   procedure load_program (R : in out robot_type; Filename : in String);
   
   -- Writes the robot program currently in the robot memory 
   -- to the named file
   procedure store_program (R : in robot_type; Filename : in String);
   
   -- Combines the fetch and execute procedures to make the robot
   -- execute the next instruction.
   procedure fetch_and_execute(R : in out robot_type; Dynamic_Behavior : in Dynamic_Behavior_Pointer);
   
   -- Returns true if last instruction was stop or robot has moved outside the the room
   function  end_of_program (R : robot_type) return Boolean;
   
   -- Robot state inquiry/modifier procedures
   -- For robot's location 
   procedure get_location (R : in     robot_type; X : out Natural; Y : out Natural);
   procedure set_location (R : in out robot_type; X : in  Positive; Y : in  Positive; 
      room : in RoomADT.room_type_ptr);
   
   -- For robot's arms
   procedure get_arms (R : in     robot_type; state : out arm_type);
   procedure set_arms (R : in out robot_type; state : in  arm_type);
   
   -- For robot's light
   procedure get_light (R : in     robot_type; state : out light_type);
   procedure set_light (R : in out robot_type; state : in  light_type);
   
   -- For robot's direction
   procedure get_direction (R : in     robot_type; state : out direction_type);
   procedure set_direction (R : in out robot_type; state : in  direction_type);
 
   -- For robot's program counter 
   procedure get_PC (R : in     robot_type; state : out address_subtype);
   procedure set_PC (R : in out robot_type; state : in  address_subtype);
 
   -- For robot's memory 
   function  get_address (R : in robot_type; loc : address_subtype) return instruction_subtype;
   function  get_address (R : in robot_type; loc : address_subtype) return String;

   -- raises an instruction error on an attempt to insert an invalid instruction into memory
   procedure set_address (R : in out robot_type; loc : address_subtype; 
      state : in instruction_subtype);
 
   -- insert/delete instruction
   procedure Insert_Instruction(R : in out robot_type);
   procedure Delete_Instruction(R : in out robot_type);
   
   -- OPCODE CONSTANTS
   stepcode : constant instruction_subtype :=   0;
   turncode : constant instruction_subtype :=  32;
   raiscode : constant instruction_subtype :=  64;
   lowrcode : constant instruction_subtype :=  96;
   senscode : constant instruction_subtype := 128;
   gotocode : constant instruction_subtype := 160;
   litecode : constant instruction_subtype := 192;
   stopcode : constant instruction_subtype := 224;
   oprdcode : constant instruction_subtype :=  31;
   
   PRIVATE
   -- INTERNAL TYPES
   -- opcodes
   type opcode_type is (stepi, turni, raisei, loweri, sensei, gotoi, lighti, stopi, noopi);

   -- instruction (operand + opcode)
   type inst_rec_type      is record
      opcode             : opcode_type := noopi;
      operand            : address_subtype := 0;
   end record;

   -- robot memory   
   type memory_type        is array (0..address_subtype'last) of inst_rec_type;

   
   -- Robot
   --    (1)  Initially the robot faces north, arms down, and its light off
   --    (2)  When the robot's light is on, it ignores all command, except light
   type robot_type is record
      memory    : memory_type;
      X, Y      : Natural         := 1;
      arms      : arm_type        := down;
      light     : light_type      := off;
      direction : direction_type  := north;
      pc        : address_subtype := 0;
      room      : RoomADT.room_type_ptr;
   end record;

   -- INTERNAL PROCEDURES
   
   -- Step causes the robot to take one step forward, if possible.  If the robot
   -- cannot step forward, it remains where it is.  
   -- The robot cannot step forward if:
   --    (1) it is next to a wall and facing the wall
   --    (2) it is one step from the wall and facing the wall with its arms raised
   procedure step (R : in out robot_type);
   
   -- Turn causes the robot to turn to the right 90 degrees, if possible.
   -- The robot cannot turn if:
   --     (1) it has a wall on its immediate right and its arms are raised
   procedure turn (R : in out robot_type);
   
   -- Raise_arms causes the robot to raise its arms, if possible.  If the robot
   -- cannot raise its arms, then it turns on its warning light.
   -- The robot cannot raise its arms if:
   --    (1) it is next to a wall and facing the wall
   procedure raise_arms (R : in out robot_type);
   
   -- Lower_arms causes the robot to lower its arms.
   procedure lower_arms (R : in out robot_type);
   
   -- Sense causes the robot to discover a wall one step in front of the robot,
   -- if the robot has its arms raised.  If a wall is discovered, then the robot
   -- turns on its warning light.
   procedure sense (R : in out robot_type; Dynamic_Behavior : in Dynamic_Behavior_Pointer);
   
   -- Light causes the robot to turn off its warning light.
   procedure light (R : in out robot_type);
   
   -- Jump sets the robot's program counter to the operand value
   -- in the gotoi command
   procedure jump (R : in out robot_type);

   -- Instruction returns either the bit string or character string represented by
   -- the opcode and operand
   function instruction (opcode : opcode_type; operand : address_subtype) return instruction_subtype; 
   function instruction (opcode : opcode_type; operand : address_subtype) return String; 

end robotADT;